import argparse
import asyncio
import logging
import os
from typing import Any, Callable
from uuid import uuid4

import aiohttp

from eyepop import EyePopSdk
from eyepop.data.data_endpoint import DataEndpoint
from eyepop.data.data_jobs import DataJob
from eyepop.data.data_types import (
    Asset,
    AutoAnnotate,
    AutoAnnotateStatus,
    Dataset,
    DatasetAutoAnnotateCreate,
    DatasetCreate,
    PredictedClass,
    Prediction,
    UserReview,
)

logging.getLogger('eyepop').setLevel(logging.DEBUG)
logging.getLogger('eyepop.requests').setLevel(logging.DEBUG)


log = logging.getLogger(__name__)

class LocalAsset(Callable[[], Any]):
    def __init__(self, path: str) -> None:
        self.path = path
        self.file = None

    def fini(self):
        if self.file is not None:
            self.file.close()
            self.file = None

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.file is not None:
            self.file.close()
            self.file = None
        self.file = open(self.path, 'rb')
        return self.file


async def import_assets(endpoint: DataEndpoint, dataset: Dataset, assets_paths: str) -> list[Asset]:
    jobs = []
    assets = []
    async def on_ready(_job: DataJob, _local_asset: LocalAsset):
        nonlocal assets
        assets.append(await _job.result())
        _local_asset.fini()

    for asset_path in os.listdir(assets_paths):
        full_path = os.path.join(assets_paths, asset_path)
        if os.path.isfile(full_path):
            log.debug("importing asset: %s", asset_path)
            local_asset = LocalAsset(full_path)
            job = await endpoint.upload_asset_job(
                local_asset,
                dataset_uuid=dataset.uuid,
                mime_type='application/octet-stream',
                sync_transform=True,
                timeout=aiohttp.ClientTimeout(total=60*60)
            )
            jobs.append(on_ready(job, local_asset))
    await asyncio.gather(*jobs)
    log.debug("imported %d assets to %s", len(assets), dataset.uuid)
    return assets

async def auto_annotate_with_test_classes(
        endpoint: DataEndpoint,
        dataset: Dataset,
        auto_annotate: AutoAnnotate,
        source: str
) -> int:
    assets = await endpoint.list_assets(dataset_uuid=dataset.uuid)
    for asset in assets:
        await endpoint.add_asset_annotation(
            asset_uuid=asset.uuid,
            auto_annotate=auto_annotate,
            source=source,
            predictions=(Prediction(
                source_width=1.0,
                source_height=1.0,
                classes=[PredictedClass(
                    classLabel="test_class"
                )]
            ),),
        )
    return len(assets)

async def approve_auto_annotate(
        endpoint: DataEndpoint,
        dataset: Dataset,
        auto_annotate: AutoAnnotate,
        source: str
) -> int:
    assets = await endpoint.list_assets(dataset_uuid=dataset.uuid)
    for asset in assets:
        await endpoint.update_asset_annotation_approval(
            asset_uuid=asset.uuid,
            auto_annotate=auto_annotate,
            source=source,
            user_review=UserReview.approved,
        )
    return len(assets)

parser = argparse.ArgumentParser(
                    prog='import dataset',
                    description='Importing local assets into a dataset',
                    epilog='.')
parser.add_argument('assets_path',)
parser.add_argument('-r', '--remove', required=False, help="leave no trace, remove the dataset at the end", default=False, action="store_true")
parser.add_argument('-d', '--dataset-uuid', required=False, help="use this dataset uuid", default=None, type=str)
main_args = parser.parse_args()

async def main():
    async with EyePopSdk.dataEndpoint(is_async=True, job_queue_length=10) as endpoint:
        try:
            if main_args.dataset_uuid is None:
                dataset = await endpoint.create_dataset(DatasetCreate(name="test import dataset"))
                log.debug("using newly created dataset: %s", dataset.uuid)
            else:
                dataset = await endpoint.get_dataset(main_args.dataset_uuid)
                log.debug("using existing dataset: %s", dataset.uuid)
            await import_assets(endpoint, dataset, main_args.assets_path)

            source = f"ep_evaluate:{uuid4().hex}"

            await endpoint.create_dataset_auto_annotate(
                auto_annotate_create=DatasetAutoAnnotateCreate(
                    auto_annotate="ep_evaluate",
                    source=source,
                    status=AutoAnnotateStatus.in_progress,
                ),
                dataset_uuid=dataset.uuid,
            )
            log.debug("auto annotate created for dataset: %s", dataset.uuid)
            num_assets = await auto_annotate_with_test_classes(
                endpoint=endpoint,
                dataset=dataset,
                auto_annotate="ep_evaluate",
                source=source,
            )
            log.debug("auto annotated %d assets in dataset: %s", num_assets, dataset.uuid)
            num_assets = await approve_auto_annotate(
                endpoint=endpoint,
                dataset=dataset,
                auto_annotate="ep_evaluate",
                source=source,
            )
            log.debug("approved %d asset annotations in dataset: %s", num_assets, dataset.uuid)
        finally:
            if main_args.dataset_uuid is None and main_args.remove:
                log.debug("removing dataset: %s", dataset.uuid)
                await endpoint.delete_dataset(dataset.uuid)

asyncio.run(main())



import logging
import logging.config
import os

# Libraries should be silent by default. Only configure console output
# if the user explicitly opts in via EYEPOP_LOG_LEVEL or LOG_LEVEL.
_explicit_level = os.environ.get('EYEPOP_LOG_LEVEL', os.environ.get('LOG_LEVEL', '')).upper()

# Always add a NullHandler to prevent "No handlers could be found" warnings
logging.getLogger('eyepop').addHandler(logging.NullHandler())

if _explicit_level:
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'default',
                'stream': 'ext://sys.stdout'
            }
        },
        'loggers': {
            'eyepop': {
                'level': _explicit_level,
                'handlers': ['console'],
                'propagate': False
            },
            'eyepop.requests': {
                'level': 'DEBUG' if _explicit_level == 'DEBUG' else 'WARNING',
                'handlers': ['console'],
                'propagate': False
            },
            'eyepop.metrics': {
                'level': _explicit_level,
                'handlers': ['console'],
                'propagate': False
            },
            'eyepop.tracer': {
                'level': _explicit_level,
                'handlers': ['console'],
                'propagate': False
            },
            'eyepop.compute': {
                'level': _explicit_level,
                'handlers': ['console'],
                'propagate': False
            }
        }
    })



import asyncio
import logging

import aiohttp

from eyepop.compute.context import ComputeContext, PipelineStatus
from eyepop.compute.responses import ComputeApiSessionResponse
from eyepop.exceptions import ComputeHealthCheckException

log = logging.getLogger("eyepop.compute")


async def wait_for_session(
    compute_config: ComputeContext, client_session: aiohttp.ClientSession
) -> bool:
    timeout = compute_config.wait_for_session_timeout
    interval = compute_config.wait_for_session_interval

    # Session endpoint health check ALWAYS uses the JWT access_token
    if not compute_config.m2m_access_token or len(compute_config.m2m_access_token.strip()) == 0:
        raise ComputeHealthCheckException(
            "No access_token in compute_config. "
            "Cannot perform session health check. "
            "This should never happen - fetch_new_compute_session should have set it.",
            session_endpoint=compute_config.session_endpoint,
        )

    auth_header = f"Bearer {compute_config.m2m_access_token}"
    log.debug(
        f"Using JWT access_token for session health check at {compute_config.session_endpoint}/health"
    )

    headers = {
        "Authorization": auth_header,
        "Accept": "application/json",
    }

    health_url = f"{compute_config.session_endpoint}/health"
    log.debug(f"Waiting for session to be ready at: {health_url}")
    log.debug(f"Timeout: {timeout}s, Interval: {interval}s")

    end_time = asyncio.get_event_loop().time() + timeout
    last_message = "No message received"
    attempt = 0

    while asyncio.get_event_loop().time() < end_time:
        attempt += 1
        try:
            log.debug(f"Health check attempt {attempt}")

            async with client_session.get(health_url, headers=headers) as response:
                log.debug(f"Health check response status: {response.status}")

                if response.status == 200:
                    log.debug("Session is ready (status 200)")
                    return True

                if response.status != 200:
                    last_message = f"Health check returned status {response.status}"
                    log.debug(last_message)
                    await asyncio.sleep(interval)
                    continue

                session_response = ComputeApiSessionResponse(**(await response.json()))
                status = session_response.session_status

                if status == PipelineStatus.RUNNING:
                    log.debug("Session is running")
                    return True
                elif status == PipelineStatus.PENDING:
                    last_message = f"Session status: {status.value}"
                    log.debug(f"Session still pending/creating: {last_message}")
                    await asyncio.sleep(interval)
                    continue
                elif status in [
                    PipelineStatus.FAILED,
                    PipelineStatus.ERROR,
                    PipelineStatus.STOPPED,
                ]:
                    raise ComputeHealthCheckException(
                        f"Session in terminal state: {status.value}. Message: {session_response.session_message}",
                        session_endpoint=compute_config.session_endpoint,
                        last_status=status.value,
                    )
                else:
                    last_message = f"Session status: {status.value}"
                    log.debug(f"Unknown session status, continuing to wait: {last_message}")
                    await asyncio.sleep(interval)
                    continue

        except ComputeHealthCheckException:
            raise
        except aiohttp.ClientResponseError as e:
            last_message = f"HTTP {e.status}: {e.message}"
            log.debug(f"HTTP error during health check: {last_message}")
        except Exception as e:
            last_message = str(e)
            log.debug(f"Exception during health check: {last_message}")

        await asyncio.sleep(interval)

    log.error(f"Session timed out after {timeout}s. Last message: {last_message}")
    raise TimeoutError(f"Session timed out after {timeout}s. Last message: {last_message}")



from typing import List

from pydantic import BaseModel, Field

from eyepop.data.types.common import Contour, Mask, Point2d, Point3d


class PredictedClass(BaseModel):
    id: int | None = None
    confidence: float | None = None
    classLabel: str
    category: str | None = None


class PredictedEmbedding(BaseModel):
    x: float | None = None
    y: float | None = None
    embedding: List[float]
    category: str | None = None


class PredictedText(BaseModel):
    id: int | None = None
    confidence: float | None = None
    text: str
    category: str | None = None


class PredictedMesh(BaseModel):
    id: int | None = None
    category: str | None = None
    confidence: float = None  # type: ignore[assignment]
    points: List[Point3d]


class PredictedKeyPoint(Point3d):
    id: int | None = None
    confidence: float | None = None
    classLabel: str | None = None
    category: str | None = None
    visible: bool | None = None


class PredictedKeyPoints(BaseModel):
    category: str | None = None
    type: str | None = None
    points: List[PredictedKeyPoint]


class PredictedObject(PredictedClass):
    trackId: int | None = None
    x: float
    y: float
    width: float
    height: float
    orientation: float | None = None
    outline: List[Point2d] | None = None
    contours: List[Contour] | None = None
    mask: Mask | None = None
    objects: List["PredictedObject"] | None = None
    classes: List[PredictedClass] | None = None
    texts: List[PredictedText] | None = None
    meshs: List[PredictedMesh] | None = None
    keyPoints: List[PredictedKeyPoints] | None = None


class PredictedMotion(BaseModel):
    begin_timestamp: int
    finished_timestamp: int | None = None


class Prediction(BaseModel):
    """Represents a prediction for an asset or a chunk of an asset."""

    source_width: float = Field(
        description="The width of the source coordinate system for all prediction coordinates."
    )
    source_height: float = Field(
        description="The height of the source coordinate system for all prediction coordinates."
    )
    timestamp: int | None = Field(
        default=None,
        description="Temporal offset of prediction from start of the media (video or audio) in nano seconds.",
    )
    captured_at: int | None = Field(
        default=None,
        description="Real time when the media was captured as epoch timestamp in nano seconds. "
        "Only provided if source provides this timestamp e.g. as timestamp/x-ntp in RTSP.",
    )
    duration: int | None = Field(
        default=None,
        description="Temporal length of the chunk that was the source for this prediction in nano seconds.",
    )
    offset: int | None = Field(
        default=None,
        description="A media specific offset. For video frames, this is the frame number of prediction. "
        "For audio samples, this is the offset of the first sample for this prediction.",
    )
    offset_duration: int | None = Field(
        default=None,
        description="Offset length of the chunk used for this prediction. It has the same format as offset.",
    )

    source_id: str | None = Field(
        default=None,
        description="Identifier for the source media (asset UUID, URL, or filename)",
    )
    seconds: float | None = Field(
        default=None,
        description="Temporal offset in seconds (convenience field, equivalent to timestamp / 1e9)",
    )

    objects: List[PredictedObject] | None = None
    classes: List[PredictedClass] | None = None
    texts: List[PredictedText] | None = None
    meshs: List[PredictedMesh] | None = None
    keyPoints: List[PredictedKeyPoints] | None = None
    embeddings: List[PredictedEmbedding] | None = None
    motions: List[PredictedMotion] | None = None



import unittest

from eyepop.data.data_types import PredictedClass, Prediction


class TestPredictionMetadata(unittest.TestCase):
    """Test new metadata fields: source_id and seconds."""

    def test_prediction_with_new_fields(self):
        """Test Prediction model with source_id and seconds."""
        pred = Prediction(
            source_width=1920.0,
            source_height=1080.0,
            timestamp=1000000000,  # 1 second in nanoseconds
            source_id="test-asset-uuid",
            seconds=1.0,
            classes=[PredictedClass(classLabel="Exterior")]
        )

        assert pred.source_id == "test-asset-uuid"
        assert pred.seconds == 1.0
        assert pred.timestamp == 1000000000

    def test_prediction_without_new_fields(self):
        """Test backward compatibility - new fields are optional."""
        pred = Prediction(
            source_width=1920.0,
            source_height=1080.0
        )

        assert pred.source_id is None
        assert pred.seconds is None

    def test_prediction_serialization(self):
        """Test JSON serialization with new fields."""
        pred = Prediction(
            source_width=612.0,
            source_height=408.0,
            timestamp=0,
            seconds=0.0,
            source_id="bb97f3ec-f6f5-11f0-9250-5ecb5ad6a98e",
            classes=[PredictedClass(classLabel="Exterior")]
        )

        json_data = pred.model_dump(exclude_none=True)
        assert json_data["source_id"] == "bb97f3ec-f6f5-11f0-9250-5ecb5ad6a98e"
        assert json_data["seconds"] == 0.0
        assert json_data["timestamp"] == 0

    def test_prediction_with_url_as_source_id(self):
        """Test source_id with URL."""
        pred = Prediction(
            source_width=1024.0,
            source_height=768.0,
            source_id="https://example.com/image.jpg",
            seconds=0.0
        )

        assert pred.source_id == "https://example.com/image.jpg"

    def test_prediction_video_frame_seconds(self):
        """Test seconds field for video frames."""
        # Frame 0 at time 0
        pred0 = Prediction(
            source_width=1920.0,
            source_height=1080.0,
            timestamp=0,
            seconds=0.0,
            offset=0
        )
        assert pred0.seconds == 0.0

        # Frame 1 at 30fps = 33.33ms = 0.0333 seconds
        pred1 = Prediction(
            source_width=1920.0,
            source_height=1080.0,
            timestamp=33333333,  # ~33ms in nanoseconds
            seconds=0.033,
            offset=1
        )
        assert abs(pred1.seconds - 0.033) < 0.001  # Allow small float precision error


if __name__ == "__main__":
    unittest.main()



from typing import Any, Callable

import aiohttp

from eyepop.client_session import ClientSession


class WorkerClientSession(ClientSession):
    async def pipeline_get(
            self, url_path_and_query: str,
            accept: str | None = None,
            timeout: aiohttp.ClientTimeout | None = None
    ) -> aiohttp.ClientResponse:
        raise NotImplementedError

    async def pipeline_post(
            self, url_path_and_query: str,
            accept: str | None = None,
            open_data: Callable | None = None,
            content_type: str | None = None,
            timeout: aiohttp.ClientTimeout | None = None
    ) -> aiohttp.ClientResponse:
        raise NotImplementedError

    async def pipeline_patch(
            self, url_path_and_query: str,
            accept: str | None = None,
            data: Any = None,
            content_type: str | None = None,
            timeout: aiohttp.ClientTimeout | None = None
    ) -> aiohttp.ClientResponse:
        raise NotImplementedError

    async def pipeline_delete(
            self, url_path_and_query: str,
            timeout: aiohttp.ClientTimeout | None = None
    ) -> aiohttp.ClientResponse:
        raise NotImplementedError



import logging
import time
from io import StringIO
from typing import Any, BinaryIO, Callable
from urllib.parse import urljoin

import aiohttp

from eyepop.compute.api import fetch_session_endpoint
from eyepop.data.types.asset import Area
from eyepop.endpoint import Endpoint
from eyepop.exceptions import (
    PopConfigurationException,
    PopNotReachableException,
    PopNotStartedException,
)
from eyepop.settings import settings
from eyepop.worker.load_balancer import EndpointLoadBalancer
from eyepop.worker.worker_client_session import WorkerClientSession
from eyepop.worker.worker_jobs import (
    WorkerJob,
    _LoadFromAssetUuidJob,
    _LoadFromJob,
    _UploadFileJob,
    _UploadStreamJob,
)
from eyepop.worker.worker_syncify import SyncWorkerJob
from eyepop.worker.worker_types import ComponentParams, MotionDetectConfig, Pop, VideoMode

log = logging.getLogger('eyepop')
log_requests = logging.getLogger('eyepop.requests')
log_metrics = logging.getLogger('eyepop.metrics')

def should_use_compute_api(pop_id: str, api_key: str | None) -> bool:
    """Determine if we should use Compute API based on pop_id and api_key."""
    if not api_key:
        return False

    is_transient = pop_id == "transient" or not pop_id

    if not is_transient:
        log.debug(f"Pop ID {pop_id} is not transient, will not use compute API")
        return False

    log.debug("Using compute API")
    return True

class WorkerEndpoint(Endpoint, WorkerClientSession):
    """Endpoint to an EyePop.ai worker."""

    def __init__(
            self,
            secret_key: str | None,
            access_token: str | None,
            api_key: str | None,
            eyepop_url: str,
            pop_id: str,
            session_uuid: str | None,
            auto_start: bool,
            stop_jobs: bool,
            job_queue_length: int,
            request_tracer_max_buffer: int,
            dataset_uuid: str | None = None,
            pipeline_image: str | None = None,
            pipeline_version: str | None = None,
    ):
        super().__init__(
            secret_key=secret_key,
            access_token=access_token,
            eyepop_url=eyepop_url,
            job_queue_length=job_queue_length,
            request_tracer_max_buffer=request_tracer_max_buffer,
            api_key=api_key,
            session_uuid=session_uuid,
        )
        self.pop_id = pop_id
        self.auto_start = auto_start
        self.stop_jobs = stop_jobs
        self.dataset_uuid = dataset_uuid

        if self.compute_ctx:
            if pipeline_image:
                self.compute_ctx.pipeline_image = pipeline_image
            if pipeline_version:
                self.compute_ctx.pipeline_version = pipeline_version
            self.is_dev_mode = not bool(session_uuid)
        else:
            self.is_dev_mode = True

        self.worker_config = None
        self.last_fetch_config_success_time = None
        self.last_fetch_config_error = None
        self.last_fetch_config_error_time = None

        # new way
        self.pop = Pop(components=[])

        self.add_retry_handler(404, self._retry_404)

    async def _retry_404(self, status_code: int, failed_attempts: int) -> bool:
        if failed_attempts > 1:
            return False
        else:
            log_requests.debug('after 404, about to retry with fresh config')
            self.worker_config = None
            return True

    async def _disconnect(self, timeout: float | None = None):
        client_timeout = None
        if timeout is not None:
            client_timeout = aiohttp.ClientTimeout(total=timeout)
        if (self.is_dev_mode and self.pop_id == 'transient'
                and self.worker_config is not None and self.worker_config.get('pipeline_id') is not None
                and self.client_session is not None):
            try:
                base_url = await self.dev_mode_base_url()
                delete_pipeline_url = f'{base_url}/pipelines/{self.worker_config["pipeline_id"]}'
                headers = {}
                authorization_header = await self._authorization_header()
                if authorization_header is not None:
                    headers['Authorization'] = authorization_header
                await self.client_session.delete(delete_pipeline_url, headers=headers, timeout=client_timeout)
            except Exception as e:
                log.exception(e, exc_info=True)
            finally:
                del self.worker_config["pipeline_id"]

    async def _reconnect(self):
        # Narrow Optional[ClientSession] — _reconnect is only called after connect()
        assert self.client_session is not None
        if self.worker_config is not None:
            return

        if self.last_fetch_config_error_time is not None and self.last_fetch_config_error_time > time.time() - settings.min_config_reconnect_secs:
            if self.last_fetch_config_error is not None:
                raise self.last_fetch_config_error
            raise aiohttp.ClientConnectionError()

        if self.last_fetch_config_success_time is not None and self.last_fetch_config_success_time > time.time() - settings.min_config_reconnect_secs:
            raise aiohttp.ClientConnectionError()

        if self.compute_ctx:
            if not self.compute_ctx.session_endpoint:
                log_requests.debug("Fetching compute API session")
                self.compute_ctx = await fetch_session_endpoint(self.compute_ctx, self.client_session)
                self.eyepop_url = self.compute_ctx.session_endpoint
                log_requests.debug(f"Compute session ready: {self.compute_ctx.session_endpoint}")

            self.worker_config = {
                "session_endpoint": self.compute_ctx.session_endpoint,
                "pipeline_id": self.compute_ctx.pipeline_id,
                "endpoints": [],
            }

            self.last_fetch_config_success_time = time.time()
            self.last_fetch_config_error = None
            self.last_fetch_config_error_time = None
        else:
            if self.pop_id == 'transient':
                config_url = f'{self.eyepop_url}/workers/config'
            else:
                config_url = f'{self.eyepop_url}/pops/{self.pop_id}/config?auto_start={self.auto_start}'
            headers = {}
            authorization_header = await self._authorization_header()
            if authorization_header is not None:
                headers['Authorization'] = authorization_header
            try:
                async with self.client_session.get(config_url, headers=headers) as response:
                    self.worker_config = await response.json()
                self.last_fetch_config_success_time = time.time()
                self.last_fetch_config_error = None
                self.last_fetch_config_error_time = None
            except aiohttp.ClientResponseError as e:
                if e.status != 401:
                    self.last_fetch_config_error = e
                    self.last_fetch_config_error_time = time.time()
                    raise e
                else:
                    self.token = None
                    self.expire_token_time = None
                    headers = {}
                    authorization_header = await self._authorization_header()
                    if authorization_header is not None:
                        headers['Authorization'] = authorization_header
                    async with self.client_session.get(config_url, headers=headers) as retried_response:
                        self.worker_config = await retried_response.json()
            except aiohttp.ClientConnectionError as e:
                self.last_fetch_config_error = e
                self.last_fetch_config_error_time = time.time()
                raise e

        if self.compute_ctx:
            log_requests.debug(f'Using compute context config: {self.worker_config}')

        base_url = await self.dev_mode_base_url()

        if self.worker_config.get('status') == 'active_prod':
            self.is_dev_mode = False

        if self.is_dev_mode:
            start_pipeline_url = f'{base_url}/pipelines'
            body = {
                "pop": self.pop.model_dump() if self.pop else {},
                "source": {
                    "sourceType": "NONE"
                },
                "idleTimeoutSeconds": 300,
                "logging": ["out_meta"],
                "videoOutput": "no_output",
            }
            if self.dataset_uuid is not None:
                body["datasetUuid"] = self.dataset_uuid

            headers = {}
            authorization_header = await self._authorization_header()
            if authorization_header is not None:
                headers['Authorization'] = authorization_header
            async with self.client_session.post(start_pipeline_url, headers=headers, json=body) as response:
                response_json = await response.json()
            self.worker_config['pipeline_id'] = response_json['id']

            if self.compute_ctx:
                self.compute_ctx.pipeline_uuid = response_json['id']
                log.debug(f"Created pipeline with ID: {response_json['id']}")

            has_base_url = ('base_url' in self.worker_config and self.worker_config['base_url'] is not None) or \
                          ('session_endpoint' in self.worker_config and self.worker_config['session_endpoint'] is not None)
            has_pipeline_id = self.worker_config.get('pipeline_id') is not None
            if not has_base_url or not has_pipeline_id:
                raise PopNotStartedException(pop_id=self.pop_id)

        if self.is_dev_mode and self.stop_jobs:
            stop_jobs_url = f'{await self.dev_mode_pipeline_base_url()}/source?mode=preempt&processing=sync'
            body = {'sourceType': 'NONE'}
            headers = {}
            authorization_header = await self._authorization_header()
            if authorization_header is not None:
                headers['Authorization'] = authorization_header
            async with self.client_session.patch(stop_jobs_url, headers=headers, json=body) as response:
                pass

        if self.is_dev_mode and self.pop is None:
            get_url = await self.dev_mode_pipeline_base_url()
            headers = {}
            authorization_header = await self._authorization_header()
            if authorization_header is not None:
                headers['Authorization'] = authorization_header
            async with self.client_session.get(get_url, headers=headers) as response:
                response_json = await response.json()
                pop_as_dict = response_json.get('pop')
                if pop_as_dict is None:
                    self.pop = None
                else:
                    self.pop = Pop(**pop_as_dict)
            log.debug('current pop is %s', self.pop)

        if self.is_dev_mode:
            if 'session_endpoint' in self.worker_config:
                base_url = self.worker_config['session_endpoint'].rstrip("/")
            else:
                base_url = urljoin(self.eyepop_url, self.worker_config['base_url']).rstrip("/")
            endpoint = {'base_url': base_url, 'pipeline_id': self.worker_config['pipeline_id']}
            self.load_balancer = EndpointLoadBalancer([endpoint])
            log.debug(f"Initialized load balancer with endpoint: {endpoint}")
        else:
            if 'session_endpoint' in self.worker_config:
                base_url = self.worker_config['session_endpoint'].rstrip("/")
                endpoint = {'base_url': base_url, 'pipeline_id': self.worker_config['pipeline_id']}
                self.load_balancer = EndpointLoadBalancer([endpoint])
                log.debug(f"Initialized load balancer with endpoint: {endpoint}")
            else:
                self.load_balancer = EndpointLoadBalancer(self.worker_config['endpoints'])
                log.debug(f"Initialized load balancer with endpoints: {self.worker_config['endpoints']}")


    # Parent returns dict | None — must match and guard against None before subscripting
    async def session(self) -> dict | None:
        session = await super().session()
        if session is None:
            return None
        session['popId'] = self.pop_id
        if self.worker_config:
            if 'session_endpoint' in self.worker_config:
                session['baseUrl'] = self.worker_config['session_endpoint']
            elif 'base_url' in self.worker_config:
                session['baseUrl'] = self.worker_config['base_url']
            session['pipelineId'] = self.worker_config['pipeline_id']
        return session

    async def get_pop(self) -> Pop | None:
        return self.pop

    async def set_pop(self, pop: Pop):
        if not self.is_dev_mode:
            raise PopConfigurationException(self.pop_id, 'set_pop not supported in production mode')
        response = await self.pipeline_patch('pop', content_type='application/json',
                                             data=pop.model_dump_json())
        self.pop = pop
        return response

    async def dev_mode_pipeline_base_url(self) -> str:
        # Was `else: pass` which made return type str | None — raise explicitly instead
        if not self.is_dev_mode:
            raise PopNotStartedException(pop_id=self.pop_id)
        base_url = await self.dev_mode_base_url()
        if self.worker_config is None:
            raise PopNotStartedException(pop_id=self.pop_id)
        return f'{base_url}/pipelines/{self.worker_config["pipeline_id"]}'

    async def upload(
            self,
            location: str,
            video_mode: VideoMode | None = None,
            params: list[ComponentParams] | None = None,
            motion_detect: MotionDetectConfig | None = None,
            roi: Area | None = None,
            fps: str | None = None,
            on_ready: Callable[[WorkerJob], None] | None = None
    ) -> WorkerJob | SyncWorkerJob:
        job = _UploadFileJob(
            location=location,
            video_mode=video_mode,
            component_params=params,
            motion_detect=motion_detect,
            roi=roi,
            fps=fps,
            session=self, on_ready=on_ready,
            callback=self.metrics_collector
        )
        await  self._task_start(job.execute())
        return job

    async def upload_stream(
            self,
            stream: BinaryIO,
            mime_type: str,
            video_mode: VideoMode | None = None,
            params: list[ComponentParams] | None = None,
            motion_detect: MotionDetectConfig | None = None,
            roi: Area | None = None,
            fps: str | None = None,
            on_ready: Callable[[WorkerJob], None] | None = None
    ) -> WorkerJob | SyncWorkerJob:
        job = _UploadStreamJob(
            stream=stream,
            mime_type=mime_type,
            video_mode=video_mode,
            component_params=params,
            motion_detect=motion_detect,
            roi=roi,
            fps=fps,
            session=self,
            on_ready=on_ready,
            callback=self.metrics_collector
        )
        await self._task_start(job.execute())
        return job

    async def load_from(
            self,
            location: str,
            params: list[ComponentParams] | None = None,
            motion_detect: MotionDetectConfig | None = None,
            roi: Area | None = None,
            fps: str | None = None,
            on_ready: Callable[[WorkerJob], None] | None = None
    ) -> WorkerJob | SyncWorkerJob:
        job = _LoadFromJob(
            location=location,
            component_params=params,
            motion_detect=motion_detect,
            roi=roi,
            fps=fps,
            session=self,
            on_ready=on_ready,
            callback=self.metrics_collector
        )
        await self._task_start(job.execute())
        return job

    async def load_asset(
            self,
            asset_uuid: str,
            params: list[ComponentParams] | None = None,
            motion_detect: MotionDetectConfig | None = None,
            roi: Area | None = None,
            fps: str | None = None,
            on_ready: Callable[[WorkerJob], None] | None = None
    ) -> WorkerJob | SyncWorkerJob:
        job = _LoadFromAssetUuidJob(
            asset_uuid=asset_uuid,
            component_params=params,
            motion_detect=motion_detect,
            roi=roi,
            fps=fps,
            session=self,
            on_ready=on_ready,
            callback=self.metrics_collector
        )
        await self._task_start(job.execute())
        return job

    async def dev_mode_base_url(self) -> str:
        if self.worker_config is None:
            await self._reconnect()
        if self.worker_config is None:
            raise PopNotStartedException(pop_id=self.pop_id)
        if 'session_endpoint' in self.worker_config:
            return self.worker_config['session_endpoint'].rstrip("/")
        return urljoin(self.eyepop_url, self.worker_config['base_url']).rstrip("/")

    #
    # Implements: _WorkerClientSession
    # Return types changed from _RequestContextManager to ClientResponse — these methods
    # await the request and return the response directly, not a context manager.
    #

    async def pipeline_get(self, url_path_and_query: str, accept: str | None = None,
                           timeout: aiohttp.ClientTimeout | None = None) -> aiohttp.ClientResponse:
        return await self._pipeline_request_with_retry('GET', url_path_and_query, accept=accept, timeout=timeout)

    async def pipeline_post(self, url_path_and_query: str, accept: str | None = None, open_data: Callable | None = None,
                            content_type: str | None = None,
                            timeout: aiohttp.ClientTimeout | None = None) -> aiohttp.ClientResponse:
        return await self._pipeline_request_with_retry('POST', url_path_and_query, accept=accept, open_data=open_data,
                                                       content_type=content_type, timeout=timeout)

    async def pipeline_patch(self, url_path_and_query: str, accept: str | None = None, data: Any = None,
                             content_type: str | None = None,
                             timeout: aiohttp.ClientTimeout | None = None) -> aiohttp.ClientResponse:
        def open_data():
            if isinstance(data, str):
                return StringIO(data)
            else:
                return data

        return await self._pipeline_request_with_retry('PATCH', url_path_and_query, accept=accept, open_data=open_data,
                                                       content_type=content_type, timeout=timeout)

    async def pipeline_delete(self, url_path_and_query: str,
                              timeout: aiohttp.ClientTimeout | None = None) -> aiohttp.ClientResponse:
        return await self._pipeline_request_with_retry('DELETE', url_path_and_query, timeout=timeout)

    async def _pipeline_request_with_retry(self, method: str, url_path_and_query: str, accept: str | None = None,
                                           open_data: Callable | None = None, content_type: str | None = None,
                                           timeout: aiohttp.ClientTimeout | None = None) -> aiohttp.ClientResponse:
        # Narrow Optional[ClientSession] for all .request() calls below
        assert self.client_session is not None
        if self.last_fetch_config_success_time is not None and self.last_fetch_config_success_time < time.time() - settings.force_refresh_config_secs:
            self.worker_config = None

        failed_attempts = 0
        retried_re_config = False
        start_time = time.time()

        while time.time() - start_time < settings.max_retry_time_secs:
            if self.worker_config is None:
                await self._reconnect()

            entry = self.load_balancer.next_entry(settings.max_retry_time_secs + 1)
            log.debug(f"Load balancer entry: {entry}")
            if entry is None:
                if not retried_re_config:
                    # pipeline might have just shut down
                    log_requests.debug('no healthy endpoints, about to retry with fresh config')
                    self.worker_config = None
                    retried_re_config = True
                    continue
                else:
                    raise PopNotReachableException(self.pop_id, self.load_balancer.get_debug_status())

            url = f'{entry.base_url}/pipelines/{entry.pipeline_id}/{url_path_and_query}'

            headers = {}
            authorization_header = await self._authorization_header()
            if authorization_header is not None:
                headers['Authorization'] = authorization_header
            if accept is not None:
                headers['Accept'] = accept
            if content_type is not None:
                headers['Content-Type'] = content_type
            try:
                if open_data is not None:
                    with open_data() as data:
                        if isinstance(data, StringIO):
                            response = await self.client_session.request(method, url, headers=headers,
                                                                         data=data.getvalue(), timeout=timeout)
                        else:
                            response = await self.client_session.request(method, url, headers=headers, data=data,
                                                                         timeout=timeout)
                else:
                    response = await self.client_session.request(method, url, headers=headers, timeout=timeout)

                entry.mark_success()

                return response
            except aiohttp.ClientResponseError as e:
                if e.status == 404:
                    # in load balanced configuration, we overwrite the standard 404 handler
                    entry.mark_error()
                else:
                    failed_attempts += 1
                    if e.status not in self.retry_handlers:
                        log_requests.exception('unexpected error: %s', e)
                        raise e
                    if not await self.retry_handlers[e.status](e.status, failed_attempts):
                        log_requests.exception('unexpected error')
                        raise e
            except aiohttp.ClientConnectionError:
                entry.mark_error()
            except Exception as e:
                log_requests.exception('unexpected error')
                raise e

        raise PopNotReachableException(self.pop_id, [])

    async def _worker_request_with_retry(self, method: str, url_path_and_query: str, accept: str | None = None,
                                         data: Any = None, content_type: str | None = None,
                                         timeout: aiohttp.ClientTimeout | None = None) -> aiohttp.ClientResponse:
        url = f'{await self.dev_mode_base_url()}/{url_path_and_query}'
        return await self.request_with_retry(method, url, accept, data, content_type, timeout)



from typing import cast

import pyarrow as pa

from eyepop.data.arrow.eyepop.predictions import (
    eyepop_predicted_classes_from_pylist,
    eyepop_predicted_embeddings_from_pylist,
    eyepop_predicted_key_pointss_from_pylist,
    eyepop_predicted_objects_from_pylist,
    eyepop_predicted_texts_from_pylist,
    eyepop_predictions_from_pylist,
    table_from_eyepop_predicted_classes,
    table_from_eyepop_predicted_embeddings,
    table_from_eyepop_predicted_key_pointss,
    table_from_eyepop_predicted_objects,
    table_from_eyepop_predicted_texts,
    table_from_eyepop_predictions,
)
from eyepop.data.arrow.schema import ANNOTATION_SCHEMA
from eyepop.data.arrow.schema_version_conversion import convert
from eyepop.data.data_types import AssetAnnotationResponse, Prediction


def table_from_eyepop_annotations(annotations: list[AssetAnnotationResponse], schema: pa.Schema = ANNOTATION_SCHEMA) -> pa.Table:
    types = []
    sources = []
    user_reviews = []
    objects = []
    classes = []
    source_model_uuids = [] if "source_model_uuid" in schema.names else None
    key_points = [] if "keyPoints" in schema.names else None
    texts = [] if "texts" in schema.names else None
    embeddings = [] if "embeddings" in schema.names else None
    timestamps = [] if "timestamp" in schema.names else None
    durations = [] if "duration" in schema.names else None
    offsets = [] if "offset" in schema.names else None
    offset_durations = [] if "offset_duration" in schema.names else None
    if "predictions" in schema.names:
        predictionss = []
        prediction_fields = cast(
            pa.StructType,
            cast(pa.ListType, schema.field("predictions").type).value_type
        ).fields
        prediction_schema = pa.schema(prediction_fields)
    else:
        predictionss = None
        prediction_fields = None
        prediction_schema = None

    for e in annotations:
        types.append(e.type)
        sources.append(e.source)
        user_reviews.append(e.user_review)
        prediction = e.predictions[0] if e.predictions else None
        if not prediction or prediction.objects is None:
            objects.append(None)
        elif len(prediction.objects) == 0:
            objects.append([])
        else:
            objects.append(table_from_eyepop_predicted_objects(
                prediction.objects,
                prediction.source_width,
                prediction.source_height,
                e.user_review,
                schema=pa.schema(schema.field(3).type.value_type), # schema for "objects" field
            ).to_struct_array())
        if not prediction or prediction.classes is None:
            classes.append(None)
        elif len(prediction.classes) == 0:
            classes.append([])
        else:
            classes.append(table_from_eyepop_predicted_classes(
                prediction.classes,
                e.user_review,
                schema=pa.schema(schema.field(4).type.value_type),  # schema for "classes" field
            ).to_struct_array())
        if source_model_uuids is not None:
            source_model_uuids.append(None)
        if key_points is not None:
            if not prediction or prediction.keyPoints is None:
                key_points.append(None)
            elif len(prediction.keyPoints) == 0:
                key_points.append([])
            else:
                key_points.append(table_from_eyepop_predicted_key_pointss(
                    prediction.keyPoints,
                    prediction.source_width,
                    prediction.source_height,
                    schema=pa.schema(schema.field(6).type.value_type),  # schema for "keyPoints" field
                ).to_struct_array())
        if texts is not None:
            if not prediction or prediction.texts is None:
                texts.append(None)
            elif len(prediction.texts) == 0:
                texts.append([])
            else:
                texts.append(table_from_eyepop_predicted_texts(prediction.texts).to_struct_array())
        if embeddings is not None:
            if not prediction or prediction.embeddings is None:
                embeddings.append(None)
            elif len(prediction.embeddings) == 0:
                embeddings.append([])
            else:
                embeddings.append(table_from_eyepop_predicted_embeddings(
                    prediction.embeddings,
                    schema=pa.schema(schema.field(8).type.value_type),  # schema for "embeddings" field
                ).to_struct_array())
        if timestamps is not None:
            timestamps.append(prediction.timestamp if prediction else None)
        if durations is not None:
            durations.append(prediction.duration if prediction else None)
        if offsets is not None:
            offsets.append(prediction.offset if prediction else None)
        if offset_durations is not None:
            offset_durations.append(prediction.offset_duration if prediction else None)
        if predictionss is not None:
            if e.predictions is not None and len(e.predictions) > 0:
                predictions = table_from_eyepop_predictions(
                    e.predictions,
                    e.user_review,
                    schema=prediction_schema,
                ).to_struct_array()
            else:
                predictions = pa.chunked_array([], type=pa.struct(prediction_fields))
            predictionss.append(predictions)


    # since 0.0
    columns = [
        pa.array(types).dictionary_encode(),
        pa.array(sources).dictionary_encode(),
        pa.array(user_reviews).dictionary_encode(),
        objects,
        classes,
    ]

    # since 1.0: source_model_uuid
    if source_model_uuids is not None:
        columns.append(pa.array(source_model_uuids).dictionary_encode())

    # since 1.1: key_points
    if key_points is not None:
        columns.append(key_points)

    # since 1.2: texts
    if texts is not None:
        columns.append(texts)

    # since 1.3: embeddings
    if embeddings is not None:
        columns.append(embeddings)

    # since 1.4: timestamp, duration, offset, offset_duration
    if timestamps is not None:
        columns.append(timestamps)
    if durations is not None:
        columns.append(durations)
    if offsets is not None:
        columns.append(offsets)
    if offset_durations is not None:
        columns.append(offset_durations)

    # since 1.7: predictions
    if predictionss is not None:
        columns.append(predictionss)

    return pa.Table.from_arrays(columns, schema=schema)


def eyepop_annotations_from_table(table: pa.Table) -> list[AssetAnnotationResponse]:
    table = convert(table, ANNOTATION_SCHEMA)
    annotations = []
    i = 0
    for batch in table.to_reader():
        types = batch.column(0).to_pylist()
        sources = batch.column(1).to_pylist()
        user_reviews = batch.column(2).to_pylist()
        objects = batch.column(3).to_pylist()
        classes = batch.column(4).to_pylist()
        # since 1.0: source_model_uuid
        source_model_uuid = batch.column(5).to_pylist()
        # since 1.1: key_points
        key_points = batch.column(6).to_pylist()
        # since 1.2: texts
        texts = batch.column(7).to_pylist()
        # since 1.3: embeddings
        embeddings = batch.column(8).to_pylist()
        # since 1.4: timestamp, duration, offset, offset_duration
        timestamps = batch.column(9).to_pylist()
        durations = batch.column(10).to_pylist()
        offsets = batch.column(11).to_pylist()
        offset_durations = batch.column(12).to_pylist()

        for j in range(len(types)):
            if objects[j] is None:
                child_objects = None
            elif len(objects[j]) == 0:
                child_objects = []
            else:
                child_objects = eyepop_predicted_objects_from_pylist(objects[j], 1.0, 1.0)

            if classes[j] is None:
                child_classes = None
            elif len(classes[j]) == 0:
                child_classes = []
            else:
                child_classes = eyepop_predicted_classes_from_pylist(classes[j])

            if key_points[j] is None:
                child_key_pointss = None
            elif len(key_points[j]) == 0:
                child_key_pointss = []
            else:
                child_key_pointss = eyepop_predicted_key_pointss_from_pylist(key_points[j], 1.0, 1.0)

            if texts[j] is None:
                child_texts = None
            elif len(texts[j]) == 0:
                child_texts = []
            else:
                child_texts = eyepop_predicted_texts_from_pylist(texts[j])

            if embeddings[j] is None:
                child_embeddings = None
            elif len(embeddings[j]) == 0:
                child_embeddings = []
            else:
                child_embeddings = eyepop_predicted_embeddings_from_pylist(embeddings[j])

            prediction = Prediction(
                source_width=1.0,
                source_height=1.0,
                timestamp=timestamps[j],
                duration=durations[j],
                offset=offsets[j],
                offset_duration=offset_durations[j],
                objects=child_objects,
                classes=child_classes,
                keyPoints=child_key_pointss,
                texts=child_texts,
                embeddings=child_embeddings,
            )

            annotations.append(AssetAnnotationResponse(
                type=types[j],
                user_review=user_reviews[j],
                source=sources[j],
                predictions=(prediction,),
                annotation=prediction,
                source_model_uuid=source_model_uuid[j],
            ))
            i += 1
    return annotations


def eyepop_annotations_from_pylist(py_list: list[dict]) -> list[AssetAnnotationResponse]:
    annotations = []
    for o in py_list:
        if o is None:
            continue
        objects = o.get('objects', None)
        if objects is None:
            child_objects = None
        else:
            child_objects = eyepop_predicted_objects_from_pylist(objects, 1.0, 1.0)
        classes = o.get('classes', None)
        if classes is None:
            child_classes = None
        else:
            child_classes = eyepop_predicted_classes_from_pylist(classes)
        key_pointss = o.get("keyPoints", None)
        if key_pointss is None:
            child_key_pointss = None
        else:
            child_key_pointss = eyepop_predicted_key_pointss_from_pylist(key_pointss, 1.0, 1.0)
        texts = o.get("texts", None)
        if texts is None:
            child_texts = None
        else:
            child_texts = eyepop_predicted_texts_from_pylist(texts)
        embeddings = o.get("embeddings", None)
        if embeddings is None:
            child_embeddings = None
        else:
            child_embeddings = eyepop_predicted_embeddings_from_pylist(embeddings)
        prediction = Prediction(
            source_width=1.0,
            source_height=1.0,
            timestamp=o.get('timestamp', None),
            duration=o.get('duration', None),
            offset=o.get('offset', None),
            offset_duration=o.get('offset_duration', None),
            objects=child_objects,
            classes=child_classes,
            keyPoints=child_key_pointss,
            texts=child_texts,
            embeddings=child_embeddings
        )
        predictions = o.get("predictions", None)
        if predictions is None:
            # backward compatible < 1.7
            eyepop_predictions = (prediction,)
        else:
            eyepop_predictions = eyepop_predictions_from_pylist(predictions)
        annotations.append(AssetAnnotationResponse(
            type=o['type'],
            user_review=o['user_review'],
            source=o['source'],
            predictions=eyepop_predictions,
            annotation=prediction,
            source_model_uuid = o.get('source_model_uuid', None)
        ))
    return annotations



from datetime import datetime
from enum import StrEnum
from typing import Annotated, Any, Literal, Sequence

from pydantic import BaseModel, Field

from eyepop.data.types.asset import Area
from eyepop.data.types.enums import EvaluationStatus, VlmAbilityStatus


class InferRuntimeConfig(BaseModel):
    """Runtime configuration for ad-hoc inference [EXPERIMENTAL].

    Common generation parameters are typed explicitly. Additional HuggingFace
    kwargs are accepted via extra="allow" and accessible via model_extra.
    """

    max_new_tokens: int | None = Field(
        default=200, description="Maximum tokens to generate"
    )
    image_size: int | None = Field(
        default=None,
        description="Target size for max dimension. 0 means auto-calculate.",
    )
    fps: float | None = Field(
        default=None,
        description="Target frames per second for sampling. 0.0 means auto-calculate.",
    )
    max_frames: int | None = Field(
        default=None, description="Maximum number of frames to load"
    )
    min_frames: int | None = Field(
        default=None, description="Minimum number of frames required"
    )
    max_aspect_ratio: float | None = Field(
        default=None, description="Maximum allowed aspect ratio"
    )
    context_length: int | None = Field(
        default=None,
        description="Maximum visual tokens to target for auto-calculation",
    )
    roi: Area | None = Field(
        default=None,
        description="Region of interest for cropping",
    )


class TransformInto(BaseModel):
    classes: Sequence[str] | None = Field(
        default=None,
        description="Transform the text output into categorical classes defined by this class label list",
    )


class InferRequest(BaseModel):
    """Client-facing ad-hoc- inference request model [EXPERIMENTAL]."""

    worker_release: str = Field(
        default="qwen3-instruct",
        description="Worker release name for routing",
    )
    text_prompt: str | None = Field(
        default=None,
        description="Prompt for the VLM",
        examples=[
            "How many distinct human beings are in the scene? Answer with an integer number"
        ],
    )
    config: InferRuntimeConfig = Field(
        default_factory=InferRuntimeConfig,
        description="Runtime configuration for inference (max_new_tokens, temperature, etc.)",
        examples=[InferRuntimeConfig(max_new_tokens=10)],
    )
    refresh: bool = Field(
        default=False,
        description="Bypass cache and force re-inference",
        examples=[False],
    )
    transform_into: TransformInto | None = Field(
        default=None,
        description="Optional instructions to transform the raw VLM text output",
    )


class EvaluateFilter(BaseModel):
    partitions: list[str] | None = Field(
        default=None,
        description="Only evaluate assets in these partitions, ignore others",
    )
    ground_truth_classes: list[str] | None = Field(
        default=None,
        description="Only evaluate assets with one of those classes in the ground truth, ignore others",
    )


class EvaluateConfig(BaseModel):
    source: str | None = Field(
        default=None,
        description="Optional source identifier to be used for the auto annotations",
    )
    dataset_uuid: str = Field(description="The Uuid dataset to evaluate.")
    filter: EvaluateFilter | None = Field(
        default=None,
        description="Only evaluate assets that matches this filter",
    )
    video_chunk_length_ns: int | None = Field(
        default=None,
        description="Video chunk length in nano seconds",
        examples=[2000000000],
    )
    video_chunk_overlap: float | None = Field(
        default=None,
        lt=0.5,
        description="Video chunk overlap ratio, possibly negative to allow gaps,"
        " e.g. -1.0 to have gaps of the same length as the chunk",
        examples=[0.1],
    )


class EvaluateRequest(EvaluateConfig):
    ability_uuid: str | None = Field(
        default=None,
        description="The uuid of the ability holding the infer config; mutually exclusive to infer",
    )
    infer: InferRequest | None = Field(
        default=None,
        description="VLM inference config; mutually exclusive to ability_uuid",
    )


class EvaluateRunInfo(BaseModel):
    """Runtime information for a competed evaluation request."""

    num_images: int = Field(description="Number of images")
    num_ground_truth_images: int = Field(
        description="Number of images with ground truth"
    )
    num_videos: int = Field(description="Number of videos")
    num_ground_truth_videos: int = Field(
        description="Number of videos with ground truth"
    )
    total_tokens: int = Field(description="Total input tokens (visual + text)")
    visual_tokens: int = Field(description="Visual tokens from all frames")
    text_tokens: int = Field(description="Text tokens from prompt")
    output_tokens: int = Field(
        description="Number of tokens generated in the model output"
    )


class EvaluateResponse(BaseModel):
    """Client-facing API response model for evaluate requests."""

    dataset_uuid: str = Field(description="The Uuid for the evaluated dataset.")
    dataset_version: int = Field(
        description="The version number of the evaluated dataset."
    )
    status: EvaluationStatus = Field(description="The final status of the evaluation.")
    status_message: str | None = Field(
        default=None, description="Optional human readable status message."
    )
    source: str = Field(
        description="Source identifier for the persisted  auto annotation."
    )
    metrics: dict[str, Any] | None = Field(default=None, description="Evaluation metrics")
    run_info: EvaluateRunInfo = Field(
        description="Runtime information, e.g. number of processed assets and used resources"
    )


class InferRunInfo(BaseModel):
    """Runtime information about the inference execution.

    Contains details about processing settings, token usage, and media characteristics.
    """

    fps: float | None = Field(
        default=None, description="Frames per second used for video processing"
    )
    image_size: int | None = Field(
        default=None, description="Maximum dimension used for image/frame resizing"
    )
    total_tokens: int | None = Field(
        default=None, description="Total input tokens (visual + text)"
    )
    visual_tokens: int | None = Field(
        default=None, description="Visual tokens from all frames"
    )
    text_tokens: int | None = Field(
        default=None, description="Text tokens from prompt"
    )
    output_tokens: int | None = Field(
        default=None, description="Number of tokens generated in the model output"
    )
    aspect_ratio: float | None = Field(
        default=None, description="Aspect ratio of the processed media (width/height)"
    )


class AutoPromptConfig(BaseModel):
    num_samples: int = Field(
        default=5,
        description="For initial auto prompt creation this is the umber of chunks to sample per class. "
                    "For auto prompt refine this is the number of failure samples to use per mismatch type (FP/FN).",
    )
    task_description: str | None = Field(
        default=None,
        description="Optional task description to append to the LLM prompt for customizing prompt creation",
    )
    infer: InferRequest = Field(
        description="InferConfig for VLM inference. The transform_into.classes field "
        "specifies the possible labels for classification.",
    )
    evaluate: EvaluateConfig = Field(
        description="EvaluateConfig for VLM evaluation.",
    )


class TaskType(StrEnum):
    base = "invalid>"
    classification = "classification"


class TaskMode(StrEnum):
    image = "image"
    video = "video"


class BaseTask(BaseModel):
    task_mode: TaskMode | None = Field(
        default=None,
        description="Defines the processing mode for the task ability when run on videos. Either on individual "
                    "images (frames of the video), or on video chunks of automatically determined length."
    )
    task_description: str | None = Field(
        default=None,
        description="Optional task description to append to the LLM prompt for customizing prompt creation",
    )


class ClassificationTask(BaseTask):
    type: Literal[TaskType.classification] = TaskType.classification
    classes: Sequence[str] = Field(
        description="List of class names."
    )
    dataset_uuid: str = Field(description="The Uuid dataset to use as input for this auto task.")
    num_samples: int = Field(
        default=5,
        description="For initial auto prompt creation this is the umber of chunks to sample per class. "
                    "For auto prompt refine this is the number of failure samples to use per mismatch type (FP/FN).",
    )


AutoTask = Annotated[ClassificationTask, Field(discriminator="type")]


class AbilityAliasEntry(BaseModel):
    alias: str
    tag: str


class VlmAbilityCreate(BaseModel):
    name: str = Field(
        description="The human readable name of the ability."
    )
    description: str = Field(
        default="",
        description="Human readable optional description of the ability."
    )
    worker_release: str = Field(
        description="The identifier of the worker release used to execute this ability."
    )
    text_prompt: str = Field(
        description="The full text prompt used for the ability."
    )
    transform_into: TransformInto = Field(
        default_factory=TransformInto,
        description="Optional transform instruction of the text result into structured response, e.g. classes.",
    )
    config: InferRuntimeConfig = Field(
        default_factory=InferRuntimeConfig,
        description="Optional inference configuration for VLM inference of this ability.",
    )
    is_public: bool = Field(
        default=False,
        description="Whether or not the ability is publicly accessible.",
    )
    video_chunk_length_ns: int | None = Field(
        default=None,
        description="Video chunk length in nano seconds",
        examples=[2000000000],
    )
    video_chunk_overlap: float | None = Field(
        default=None,
        lt=0.5,
        description="Video chunk overlap ratio, possibly negative to allow gaps,"
        " e.g. -1.0 to have gaps of the same length as the chunk",
        examples=[0.1],
    )


class VlmAbilityUpdate(BaseModel):
    name: str | None = Field(
        default=None,
        description="The human readable name of the ability."
    )
    description: str | None = Field(
        default=None,
        description="Human readable optional description of the ability."
    )
    worker_release: str | None = Field(
        default=None,
        description="The identifier of the worker release used to execute this ability."
    )
    text_prompt: str | None = Field(
        default=None,
        description="The full text prompt used for the ability."
    )
    transform_into: TransformInto | None = Field(
        default=None,
        description="Optional transform instruction of the text result into structured response, e.g. classes.",
    )
    config: InferRuntimeConfig | None = Field(
        default=None,
        description="Optional inference configuration for VLM inference of this ability.",
    )
    is_public: bool | None = Field(
        default=None,
        description="Whether or not the ability is publicly accessible.",
    )
    video_chunk_length_ns: int | None = Field(
        default=None,
        description="Video chunk length in nano seconds",
        examples=[2000000000],
    )
    video_chunk_overlap: float | None = Field(
        default=None,
        lt=1.0,
        description="Video chunk overlap ratio, possibly negative to allow gaps,"
        " e.g. -1.0 to have gaps of the same length as the chunk",
        examples=[0.1],
    )


class VlmAbilityResponse(BaseModel):
    uuid: str = Field(
        description="The uuid of the ability."
    )
    created_at: datetime | None = Field(
        default=None,
        description="The datetime when the ability was created."
    )
    updated_at: datetime | None = Field(
        default=None,
        description="The datetime when the ability was last updated."
    )
    account_uuid: str = Field(
        description="The uuid of the account associated with the ability."
    )
    status: VlmAbilityStatus = Field(
        description="The status of the ability."
    )
    vlm_ability_group_uuid: str | None = Field(
        default=None,
        description="The uuid of the VLM ability group associated with the ability."
    )
    name: str = Field(
        description="The human readable name of the ability."
    )
    description: str = Field(
        default="",
        description="Human readable optional description of the ability."
    )
    worker_release: str | None = Field(
        default=None,
        description="The identifier of the worker release used to execute this ability."
    )
    text_prompt: str | None = Field(
        default=None,
        description="The full text prompt used for the ability."
    )
    transform_into: TransformInto | None = Field(
        default=None,
        description="Optional transform instruction of the text result into structured response, e.g. classes.",
    )
    config: InferRuntimeConfig | None = Field(
        default=None,
        description="Optional inference configuration for VLM inference of this ability.",
    )
    is_public: bool = Field(
        default=False,
        description="Whether or not the ability is publicly accessible.",
    )
    video_chunk_length_ns: int | None = Field(
        default=None,
        description="Video chunk length in nano seconds",
        examples=[2000000000],
    )
    video_chunk_overlap: float | None = Field(
        default=None,
        lt=1.0,
        description="Video chunk overlap ratio, possibly negative to allow gaps,"
        " e.g. -1.0 to have gaps of the same length as the chunk",
        examples=[0.1],
    )
    alias_entries: list[AbilityAliasEntry] = Field(
        default=[],
        description="The list of alias entries assigned to this ability."
    )
    auto_prompt: AutoPromptConfig | None = Field(
        description="Optionally, the auto prompt config this ability was created or refined  with.",
        default=None,
        deprecated="Use auto_task instead",
    )
    auto_task: AutoTask | None = Field(
        description="Optionally, the auto task config this ability was created or refined  with."
                    "Replaces the deprecated auto_prompt.",
        default=None,
    )


class VlmAbilityGroupCreate(BaseModel):
    name: str = Field(
        description="Human readable name of the ability group",
    )
    description: str = Field(
        default="",
        description="Optional human readable description of the ability group",
    )

    auto_prompt: AutoPromptConfig | None = Field(
        description="Optionally create an ability in this group via auto prompt agent. "
                    "The created ability will stay in 'in_progress' state until the agent completes.",
        default=None,
        deprecated="Use auto_task instead"
    )

    auto_task: AutoTask | None = Field(
        description="Optionally create an ability in this group via auto task agent. "
                    "The created ability will stay in 'in_progress' state until the agent completes."
                    "Replaces the deprecated auto_prompt.",
        default=None,
    )

    default_alias_name: str | None = Field(
        default=None,
        description="Optionally use this name as default alias name for all abilities in this group",
    )
    default_dataset_uuid: str | None = Field(
        default=None,
        description="Optionally use this dataset UUID as evaluation target for all abilities in this group",
    )


class VlmAbilityGroupUpdate(BaseModel):
    name: str | None = Field(
        default=None,
        description="Human readable name of the ability group",
    )
    description: str | None = Field(
        default=None,
        description="Optional human readable description of the ability group",
    )
    default_alias_name: str | None = Field(
        default=None,
        description="Optionally use this name as default alias name for all abilities in this group",
    )
    default_dataset_uuid: str | None = Field(
        default=None,
        description="Optionally use this dataset UUID as evaluation target for all abilities in this group",
    )


class VlmAbilityGroupResponse(BaseModel):
    uuid: str = Field(
        description="The uuid of the ability group."
    )
    created_at: datetime | None = Field(
        default=None,
        description="The datetime when the ability group was created.",
    )
    updated_at: datetime | None = Field(
        default=None,
        description="The datetime when the ability group was last updated.",
    )
    account_uuid: str = Field(
        description="The uuid of the account associated with the ability group."
    )
    name: str = Field(
        description="Human readable name of the ability group."
    )
    description: str = Field(
        default="",
        description="Optional human readable description of the ability group",
    )
    default_alias_name: str | None = Field(
        default=None,
        description="Name as default alias name for all abilities in this group",
    )
    default_dataset_uuid: str | None = Field(
        default=None,
        description="Dataset UUID as default evaluation target for all abilities in this group",
    )



import asyncio
from contextlib import suppress


class Periodic:
    def __init__(self, func, time: float):
        self.func = func
        self.time = time
        self.is_started = False
        self._task: asyncio.Task | None = None

    async def start(self):
        if not self.is_started:
            self.is_started = True
            self._task = asyncio.create_task(self._run())
            # Yield so the task enters _run and hits its first await.
            # Without this, stop() can cancel a task whose coroutine was
            # never started, triggering "coroutine was never awaited".
            await asyncio.sleep(0)

    async def stop(self):
        if self.is_started:
            self.is_started = False
            if self._task is not None and not self._task.done():
                self._task.cancel()
                with suppress(asyncio.CancelledError):
                    await self._task
            self._task = None

    async def _run(self):
        while True:
            await asyncio.sleep(self.time)
            await self.func()



from eyepop.data.types.enums import (
    APPLICATION_JSON,
    MIME_TYPE_APACHE_ARROW_FILE,
    AnnotationInclusionMode,
    AnnotationType,
    ArgoWorkflowPhase,
    ArtifactType,
    AssetInclusionMode,
    AssetStatus,
    AssetUrlType,
    AutoAnnotate,
    AutoAnnotateStatus,
    AutoAnnotateTask,
    ChangeType,
    EvaluationStatus,
    ExportedBy,
    ModelExportFormat,
    ModelExportStatus,
    ModelStatus,
    ModelTask,
    ModelTrainingStage,
    ModelTrainingStatusCode,
    ModelType,
    TranscodeMode,
    UserReview,
    VlmAbilityStatus,
)

from eyepop.data.types.common import (
    Box,
    Contour,
    Mask,
    Point2d,
    Point3d,
)

from eyepop.data.types.prediction import (
    PredictedClass,
    PredictedEmbedding,
    PredictedKeyPoint,
    PredictedKeyPoints,
    PredictedMesh,
    PredictedObject,
    PredictedText,
    Prediction,
)

from eyepop.data.types.dataset import (
    AutoAnnotateParams,
    AutoAnnotatePrompt,
    Dataset,
    DatasetAutoAnnotate,
    DatasetAutoAnnotateCreate,
    DatasetAutoAnnotateUpdate,
    DatasetCreate,
    DatasetParent,
    DatasetResponse,
    DatasetUpdate,
    DatasetVersion,
    DatasetVersionAssetStats,
    DatasetVersionResponse,
)

from eyepop.data.types.asset import (
    Asset,
    AssetAnnotation,
    AssetAnnotationImport,
    AssetAnnotationResponse,
    AssetImport,
    AssetResponse,
    RectangleArea,
    Roi,
    TimeSpan,
)

from eyepop.data.types.model import (
    AliasResolution,
    ExportedAliasResponse,
    ExportedUrlResponse,
    Model,
    ModelAlias,
    ModelAliasCreate,
    ModelAliasResponse,
    ModelAliasUpdate,
    ModelCreate,
    ModelExport,
    ModelExportResponse,
    ModelMetrics,
    ModelResponse,
    ModelSample,
    ModelTrainingAuditRecord,
    ModelTrainingEvent,
    ModelTrainingProgress,
    ModelUpdate,
    QcAiHubExportParams,
    Tag,
    TagResponse,
)

from eyepop.data.types.events import (
    ChangeEvent,
    EventHandler,
)

from eyepop.data.types.workflow import (
    CreateWorkflowBody,
    CreateWorkflowParameters,
    CreateWorkflowResponse,
    DownloadResponse,
    ListWorkflowItem,
    ListWorkflowItemMetadata,
    ListWorkflowItemMetadataLabels,
)

from eyepop.data.types.vlm import (
    AbilityAliasEntry,
    EvaluateFilter,
    EvaluateRequest,
    EvaluateResponse,
    EvaluateRunInfo,
    InferRequest,
    InferRunInfo,
    InferRuntimeConfig,
    TransformInto,
    VlmAbilityCreate,
    VlmAbilityGroupCreate,
    VlmAbilityGroupResponse,
    VlmAbilityGroupUpdate,
    VlmAbilityResponse,
    VlmAbilityUpdate,
)

__all__ = [
    "AnnotationInclusionMode",
    "AnnotationType",
    "ArgoWorkflowPhase",
    "ArtifactType",
    "AssetInclusionMode",
    "AssetStatus",
    "AssetUrlType",
    "AutoAnnotate",
    "AutoAnnotateStatus",
    "AutoAnnotateTask",
    "ChangeType",
    "EvaluationStatus",
    "ExportedBy",
    "ModelExportFormat",
    "ModelExportStatus",
    "ModelStatus",
    "ModelTask",
    "ModelTrainingStage",
    "ModelTrainingStatusCode",
    "ModelType",
    "TranscodeMode",
    "UserReview",
    "VlmAbilityStatus",
    "APPLICATION_JSON",
    "MIME_TYPE_APACHE_ARROW_FILE",
    "Box",
    "Contour",
    "Mask",
    "Point2d",
    "Point3d",
    "PredictedClass",
    "PredictedEmbedding",
    "PredictedKeyPoint",
    "PredictedKeyPoints",
    "PredictedMesh",
    "PredictedObject",
    "PredictedText",
    "Prediction",
    "AutoAnnotateParams",
    "AutoAnnotatePrompt",
    "Dataset",
    "DatasetAutoAnnotate",
    "DatasetAutoAnnotateCreate",
    "DatasetAutoAnnotateUpdate",
    "DatasetCreate",
    "DatasetParent",
    "DatasetResponse",
    "DatasetUpdate",
    "DatasetVersion",
    "DatasetVersionAssetStats",
    "DatasetVersionResponse",
    "Asset",
    "AssetAnnotation",
    "AssetAnnotationImport",
    "AssetAnnotationResponse",
    "AssetImport",
    "AssetResponse",
    "AliasResolution",
    "RectangleArea",
    "TimeSpan",
    "Roi",
    "ExportedAliasResponse",
    "ExportedUrlResponse",
    "Model",
    "ModelAlias",
    "ModelAliasCreate",
    "ModelAliasResponse",
    "ModelAliasUpdate",
    "ModelCreate",
    "ModelExport",
    "ModelExportResponse",
    "ModelMetrics",
    "ModelResponse",
    "ModelSample",
    "ModelTrainingAuditRecord",
    "ModelTrainingEvent",
    "ModelTrainingProgress",
    "ModelUpdate",
    "QcAiHubExportParams",
    "Tag",
    "TagResponse",
    "ChangeEvent",
    "EventHandler",
    "CreateWorkflowBody",
    "CreateWorkflowParameters",
    "CreateWorkflowResponse",
    "DownloadResponse",
    "ListWorkflowItem",
    "ListWorkflowItemMetadata",
    "ListWorkflowItemMetadataLabels",
    "AbilityAliasEntry",
    "EvaluateFilter",
    "EvaluateRequest",
    "EvaluateResponse",
    "EvaluateRunInfo",
    "InferRequest",
    "InferRunInfo",
    "InferRuntimeConfig",
    "TransformInto",
    "VlmAbilityCreate",
    "VlmAbilityGroupCreate",
    "VlmAbilityGroupResponse",
    "VlmAbilityGroupUpdate",
    "VlmAbilityResponse",
    "VlmAbilityUpdate",
]



import argparse
import asyncio
import logging

import aiofiles

from eyepop import EyePopSdk
from eyepop.data.data_types import AssetUrlType, TranscodeMode

log = logging.getLogger('eyepop.example')

parser = argparse.ArgumentParser(
                    prog='download video asset',
                    description='Downloading a video asset, optionally trimmed to a certain length',
                    epilog='.')
parser.add_argument('output_path',)
parser.add_argument('-a', '--asset-uuid', required=True, help="export this asset uuid", default=None, type=str)
parser.add_argument('-s', '--start', required=False, type=float, help="start at this second", default=None)
parser.add_argument('-e', '--end', required=False, type=float, help="end at this second", default=None)
main_args = parser.parse_args()

async def main():
    async with EyePopSdk.dataEndpoint(is_async=True) as endpoint:
        result = await endpoint.download_asset(
            asset_uuid=main_args.asset_uuid,
            transcode_mode=TranscodeMode.video_original_size,
            start_timestamp=int(main_args.start * 1000 * 1000 * 1000) if main_args.start is not None else None,
            end_timestamp=int(main_args.end * 1000 * 1000 * 1000) if main_args.end is not None else None,
            url_type=AssetUrlType.gcs,
        )
        log.info("asset %s is stored in the Cloud at %s", main_args.asset_uuid, result.url)
        stream = await endpoint.download_asset(
                asset_uuid=main_args.asset_uuid,
                transcode_mode=TranscodeMode.video_original_size,
                start_timestamp=int(main_args.start * 1000 * 1000 * 1000) if main_args.start is not None else None,
                end_timestamp=int(main_args.end * 1000 * 1000 * 1000) if main_args.end is not None else None,
        )
        async with aiofiles.open(main_args.output_path, mode='wb') as out_file:
            while True:
                buffer = await stream.read(65536)
                if not buffer:
                    break
                await out_file.write(buffer)
        log.info("asset %s downloaded locally to %s", main_args.asset_uuid, main_args.output_path)

asyncio.run(main())



from typing import Awaitable, Callable

from pydantic import BaseModel

from eyepop.data.types.enums import ChangeType


class ChangeEvent(BaseModel):
    change_type: ChangeType
    account_uuid: str
    dataset_uuid: str | None
    dataset_version: int | None
    asset_uuid: str | None
    mdl_uuid: str | None
    workflow_id: str | None
    message: str | None
    workflow_task_name: str | None


EventHandler = Callable[[ChangeEvent], Awaitable[None]]



import argparse
import ast
import asyncio
import base64
import json
import logging
import os
import sys
from argparse import Namespace
from io import BytesIO
from typing import Any

from dotenv import load_dotenv
from PIL import Image
from pybars import Compiler
from pydantic import TypeAdapter
from webui import webui

from eyepop import EyePopSdk, Job
from eyepop.data.data_types import TranscodeMode
from eyepop.data.types.asset import Area, RectangleArea
from eyepop.worker.worker_types import (
    ComponentParams,
    ContourFinderComponent,
    ContourType,
    CropForward,
    ForwardComponent,
    FullForward,
    InferenceComponent,
    MotionDetectConfig,
    MotionModel,
    Pop,
    TrackingComponent,
)

load_dotenv()

logging.basicConfig(level=logging.INFO)
log = logging.getLogger('eyepop.example')

script_dir = os.path.dirname(__file__)


pop_examples = {
    "vehicles": Pop(components=[
        InferenceComponent(
            ability='eyepop.vehicle:latest',
            categoryName="vehicles",
            confidenceThreshold=0.8,
            forward=CropForward(
                targets=[TrackingComponent(
                    maxAgeSeconds=5.0,
                    motionModel=MotionModel.CONSTANT_VELOCITY,
                    agnostic=True,
                    classHysteresis=True,
                ), InferenceComponent(
                    ability='eyepop.vehicle.licence-plate:latest',
                    topK=1,
                    forward=CropForward(
                        targets=[InferenceComponent(
                            ability='eyepop.text.recognize.landscape:latest',
                            categoryName="licence-plate"
                        )]
                    )
                )]
            )
        )
    ]),

    "person": Pop(components=[
        InferenceComponent(
            ability='eyepop.person:latest',
            categoryName="person"
        )
    ]),

    "2d-body-points": Pop(components=[
        InferenceComponent(
            ability='eyepop.person:latest',
            categoryName="person",
            forward=CropForward(
                maxItems=128,
                targets=[InferenceComponent(
                    ability='eyepop.person.2d-body-points:latest',
                    categoryName="2d-body-points",
                    confidenceThreshold=0.25
                )]
        ))
    ]),

    "faces": Pop(components=[
        InferenceComponent(
            ability='eyepop.person:latest',
            categoryName="person",
            forward=CropForward(
                maxItems=128,
                targets=[InferenceComponent(
                    ability='eyepop.person.face.short-range:latest',
                    categoryName="2d-face-points",
                    forward=CropForward(
                        boxPadding=1.5,
                        orientationTargetAngle=-90.0,
                        targets=[InferenceComponent(
                            ability='eyepop.person.face-mesh:latest',
                            categoryName="3d-face-mesh"
                        )]
                    )
                )]
            )
        )
    ]),

    "hands": Pop(components=[
        InferenceComponent(
            ability='eyepop.person:latest',
            categoryName="person",
            forward=CropForward(
                maxItems=128,
                boxPadding=0.25,
                targets=[InferenceComponent(
                    ability='eyepop.person.palm:latest',
                    forward=CropForward(
                        includeClasses=["hand circumference"],
                        orientationTargetAngle=-90.0,
                        targets=[InferenceComponent(
                            ability='eyepop.person.3d-hand-points:latest',
                            categoryName="3d-hand-points"
                        )]
                    )
                )]
            )
        )
    ]),

    "3d-body-points": Pop(components=[
        InferenceComponent(
            ability='eyepop.person:latest',
            categoryName="person",
            forward=CropForward(
                boxPadding=0.5,
                targets=[InferenceComponent(
                    ability='eyepop.person.pose:latest',
                    hidden=True,
                    forward=CropForward(
                        boxPadding=0.5,
                        orientationTargetAngle=-90.0,
                        targets=[InferenceComponent(
                            ability='eyepop.person.3d-body-points.heavy:latest',
                            categoryName="3d-body-points",
                            confidenceThreshold=0.25
                        )]
                    )
                )]
            )
        )
    ]),

    "text": Pop(components=[
        InferenceComponent(
            ability='eyepop.text:latest',
            categoryName="text",
            confidenceThreshold=0.7,
            forward=CropForward(
                maxItems=128,
                targets=[InferenceComponent(
                    ability='eyepop.text.recognize.landscape:latest',
                    confidenceThreshold=0.1
                )]
            )
        )
    ]),

    "sam1": Pop(components=[
        InferenceComponent(
            ability='eyepop.sam.small:latest',
            id=1,
            forward=FullForward(
                targets=[ContourFinderComponent(
                    contourType=ContourType.POLYGON,
                    areaThreshold=0.005
                )]
            )
        )
    ]),

    "sam2": Pop(components=[
        InferenceComponent(
            ability="eyepop.sam2.encoder.tiny:latest",
            id=1,
            hidden=True,
            forward=FullForward(
                targets=[InferenceComponent(
                    ability='eyepop.sam2.decoder:latest',
                    forward=FullForward(
                        targets=[ContourFinderComponent(
                            contourType=ContourType.POLYGON,
                            areaThreshold=0.005
                        )]
                    )
                )]
            )
        )
    ]),
    "image-contents": Pop(components=[
        InferenceComponent(
            id=1,
            ability='eyepop.image-contents:latest',
        )
    ]),
    "localize-objects": Pop(components=[
        InferenceComponent(
            id=1,
            ability='eyepop.localize-objects:latest',
        )
    ]),
    "localize-objects-plus": Pop(components=[
        InferenceComponent(
            id=1,
            ability='eyepop.localize-objects:latest',
            params={
                "prompts": [{"prompt": "person"}]
            },
            forward=CropForward(
                targets=[InferenceComponent(
                    ability='eyepop.image-contents:latest',
                    params={
                        "prompts": [{"prompt": "hair color blond"},{"prompt": "hair color brown"}]
                    }
                )],
            )
        )
    ]),
    "localize-objects-t4": Pop(components=[
        InferenceComponent(
            id=1,
            ability='eyepop.localize-objects:latest',
            params={
                "prompts": [{"prompt": "person"}]
            },
            forward=CropForward(
                targets=[InferenceComponent(
                    ability='eyepop.image-contents-t4:latest',
                    params={
                        "prompts": [{"prompt": "shirt color?"}]
                    }
                )],
            )
        )
    ]),
}

def list_of_points(arg: str) -> list[dict[str, Any]]:
    points = []
    points_as_tuples = ast.literal_eval(f'[{arg}]')
    for tuple in points_as_tuples:
        points.append({
            "x": tuple[0],
            "y": tuple[1]
        })
    return points


def list_of_boxes(arg: str) -> list[dict[str, Any]]:
    boxes = []
    boxes_as_tuples = ast.literal_eval(f'[{arg}]')
    for tuple in boxes_as_tuples:
        boxes.append({
            "topLeft": {
                "x": tuple[0],
                "y": tuple[1]
            },
            "bottomRight": {
                "x": tuple[2],
                "y": tuple[3]
            }
        })
    return boxes


def rectangle_roi(arg: str) -> Area:
    roi = ast.literal_eval(arg)

    return RectangleArea(
        x=roi[0],
        y=roi[1],
        width=roi[2],
        height=roi[3],
    )


def add_optional_tracking_to_component(component: ForwardComponent, tracking_args: Namespace):
    if tracking_args.tracking:
        tracking_component = TrackingComponent()
        if tracking_args.tracking_reid_model is not None:
            tracking_component.reidModel = tracking_args.tracking_reid_model
        if tracking_args.tracking_max_age is not None:
            tracking_component.maxAgeSeconds = tracking_args.tracking_max_age
        if tracking_args.tracking_iou_threshold is not None:
            tracking_component.iouThreshold = tracking_args.tracking_iou_threshold
        if tracking_args.tracking_sim_threshold is not None:
            tracking_component.simThreshold = tracking_args.tracking_sim_threshold
        if tracking_args.tracking_agnostic is not None:
            tracking_component.agnostic = tracking_args.tracking_agnostic
        if tracking_args.tracking_motion_model is not None:
            tracking_component.motionModel = MotionModel(tracking_args.tracking_motion_model)
        component.forward = CropForward(
            targets=[tracking_component]
        )


parser = argparse.ArgumentParser(
                    prog='Pop examples',
                    description='Demonstrates the composition of a Pop',
                    epilog='.')
parser.add_argument('-d', '--dump', required=False, help="dump composable pop definition to stdout", default=False, action="store_true")
parser.add_argument('-l', '--local-path', required=False, type=str, default=False, help="run the inference on a local file, or all files on a directory")
parser.add_argument('-a', '--asset-uuid', required=False, type=str, default=False, help="run the inference on an asset by its Uuid")
parser.add_argument('-u', '--url', required=False, type=str, default=False, help="run the inference on a remote Url")
parser.add_argument('-p', '--pop', required=False, type=str, help="run this pop", choices=list(pop_examples.keys()))
parser.add_argument('-s', '--session', required=False, type=str, help="Use an existing a session uuid and do not set a pop, use the sessions as preconfigured", default=None)
parser.add_argument('-m', '--model-uuid', required=False, type=str, action="append", help="run this model(s) by uuid")
parser.add_argument('-ma', '--model-alias', required=False, type=str, action="append", help="run this model(s) by its tagged alias")
parser.add_argument('-ms1', '--model-uuid-sam1', required=False, type=str, help="run this model by uuid and compose with SAM1 (EfficientSAM) and Contour Finder")
parser.add_argument('-ms2', '--model-uuid-sam2', required=False, type=str, help="run this model by uuid and compose with SAM2 and Contour Finder")
parser.add_argument('-po', '--points', required=False, type=list_of_points, help="List of POIs as coordinates like (x1, y1), (x2, y2) in the original image coordinate system")
parser.add_argument('-bo', '--boxes', required=False, type=list_of_boxes, help="List of POIs as boxes like (left1, top1, right1, bottom1), (left1, top1, right1, bottom1) in the original image coordinate system")
parser.add_argument('-sp', '--single-prompt', required=False, type=str, help="Single prompt to pass as parameter")
parser.add_argument('-sl', '--single-label', required=False, type=str, help="Single label to use as result for single prompt to pass as parameter")
parser.add_argument('-pr', '--prompt', required=False, type=str, help="Prompt to pass as parameter", action="append")
parser.add_argument('-v', '--visualize', required=False, help="show rendered output", default=False, action="store_true")
parser.add_argument('-o', '--output', required=False, help="print results to stdout", default=False, action="store_true")
parser.add_argument('-ds', '--dataset-uuid', required=False, type=str, help="Ingest all assets into a dataset uuid", default=None)
parser.add_argument('-tk', '--top-k', required=False, type=int, help="For --model-uuid and -model-alias apply this top-k filter", default=None)
parser.add_argument('-ct', '--confidence-threshold', required=False, type=float, help="For --model-uuid and -model-alias apply this confidence threshold filter", default=None)
parser.add_argument('--fps', required=False, type=str, help="For --model-uuid and -model-alias apply this target fps - e.g. 15/1", default=None)
parser.add_argument('-vl', '--video-chunk-length-seconds', required=False, type=float, help="For --model-uuid and -model-alias apply this video chunk length", default=None)
parser.add_argument('-vo', '--video-chunk-overlap', required=False, type=float, help="For --model-uuid and -model-alias apply this video chunk overlap", default=None)

# Optional tracking for simple pops my model uuid oder model alias
parser.add_argument('--tracking', required=False, help="Track objects in videos", default=False, action="store_true")
parser.add_argument('--tracking-reid-model', required=False, help="Use re-id model uuid for tracking", default=None, type=str)
parser.add_argument('--tracking-agnostic', required=False, help="Track objects class-agnostic", default=False, action="store_true")
parser.add_argument('--tracking-max-age', required=False, help="Max age in seconds for unmatched tracks", default=None, type=float)
parser.add_argument('--tracking-iou-threshold', required=False, help="IoU threshold to match tracks", default=None, type=float)
parser.add_argument('--tracking-sim-threshold', required=False, help="Similarity threshold to match tracks by re-id", default=None, type=float)
parser.add_argument('--tracking-motion-model', required=False, help="Pick a motion model for tracking, one of 'random_walk', 'constant_velocity' or 'constant_acceleration'", default=None, type=str)

# Optional motion detection parameters
parser.add_argument('--motion-detect', required=False, help="Skip video frames w/o detected motion", default=False, action="store_true")

# Optional global ROI parameters
parser.add_argument('--roi', required=False, type=rectangle_roi, help="Rectangular ROI as (x, y, width, height)")


main_args = parser.parse_args()

if not main_args.local_path and not main_args.url and not main_args.asset_uuid:
    print("Need something to run inference on; pass either --url or --local-path or --asset-uuid")
    parser.print_help()
    sys.exit(1)

if (not main_args.pop
        and not main_args.model_uuid
        and not main_args.model_alias
        and not main_args.model_uuid_sam1
        and not main_args.model_uuid_sam2
        and not main_args.session):
    print("Need something do do, pass either --pop or --model-uuid or --model-alias or --model-uuid-sam1 "
          "or --model-uuid-sam2 (or start with a preconfigured session with --session)")
    parser.print_help()
    sys.exit(1)

if main_args.pop:
    pop = pop_examples[main_args.pop]
elif main_args.model_uuid:
    pop = Pop(components=[
        InferenceComponent(
            id=i+1,
            abilityUuid=uuid,
        ) for i, uuid in enumerate(main_args.model_uuid)
    ])
    if main_args.top_k is not None:
        for c in pop.components:
            c.topK = main_args.top_k

    if main_args.confidence_threshold is not None:
        for c in pop.components:
            c.confidenceThreshold = main_args.confidence_threshold

    if main_args.video_chunk_length_seconds is not None:
        for c in pop.components:
            c.videoChunkLengthSeconds = main_args.video_chunk_length_seconds

    if main_args.video_chunk_overlap is not None:
        for c in pop.components:
            c.videoChunkOverlap = main_args.video_chunk_overlap

    add_optional_tracking_to_component(pop.components[0], main_args)
elif main_args.model_alias:
    pop = Pop(components=[
        InferenceComponent(
            id=i+1,
            ability=alias,
        ) for i, alias in enumerate(main_args.model_alias)
    ])

    if main_args.top_k is not None:
        for c in pop.components:
            c.topK = main_args.top_k

    if main_args.confidence_threshold is not None:
        for c in pop.components:
            c.confidenceThreshold = main_args.confidence_threshold

    if main_args.video_chunk_length_seconds is not None:
        for c in pop.components:
            c.videoChunkLengthSeconds = main_args.video_chunk_length_seconds

    if main_args.video_chunk_overlap is not None:
        for c in pop.components:
            c.videoChunkOverlap = main_args.video_chunk_overlap

    add_optional_tracking_to_component(pop.components[0], main_args)
elif main_args.model_uuid_sam1:
    pop = Pop(components=[
        InferenceComponent(
            abilityUuid=main_args.model_uuid_sam1,
            forward=CropForward(
                targets=[InferenceComponent(
                    ability='eyepop.sam.small:latest',
                    forward=FullForward(
                        targets=[ContourFinderComponent(
                            contourType=ContourType.POLYGON,
                            areaThreshold=0.005
                        )]
                    )
                )]
            )
        )
    ])
elif main_args.model_uuid_sam2:
    pop = Pop(components=[
        InferenceComponent(
            ability="eyepop.sam2.encoder.tiny:latest",
            hidden=True,
            forward=FullForward(
                targets=[InferenceComponent(
                    abilityUuid=main_args.model_uuid_sam2,
                    forward=CropForward(
                        targets=[InferenceComponent(
                            ability='eyepop.sam2.decoder:latest',
                            forward=FullForward(
                                targets=[ContourFinderComponent(
                                    contourType=ContourType.POLYGON,
                                    areaThreshold=0.005
                                )]
                            )
                        )]
                    )
                )]
            )
        )
    ])
elif main_args.session:
    pop = None
else:
    raise ValueError("pop or model required (or a preconfigured session)")

params = None
if main_args.points:
    params = [
        ComponentParams(componentId=1, values={
          "roi": {
              "points": main_args.points
          }
        })
    ]
elif main_args.boxes:
    params = [
        ComponentParams(componentId=1, values={
            "roi": {
                "boxes": main_args.boxes
            }
        })
    ]
elif main_args.prompt is not None and len(main_args.prompt) > 0:
    params = [
        ComponentParams(componentId=1, values={
            "prompts": [{"prompt": p} for p in main_args.prompt]
        })
    ]
elif main_args.single_prompt is not None:
    params = [
        ComponentParams(componentId=1, values={
            "prompt": main_args.single_prompt,
            "label": main_args.single_label if main_args.single_label else main_args.single_prompt
        })
    ]

async def main(args) -> tuple[dict[str, Any] | None, str | None]:
    visualize_prediction = None
    visualize_path = None
    example_image_src = None
    motion_detect = MotionDetectConfig(motionGap=1) if args.motion_detect else None

    if args.dump and pop:
        print("Pop:", pop.model_dump_json(indent=2))
        if params:
            print("Params:", TypeAdapter(list[ComponentParams]).dump_json(params, indent=2).decode("utf-8"))

    async with EyePopSdk.async_worker(dataset_uuid=args.dataset_uuid, session_uuid=args.session) as endpoint:
        if pop:
            await endpoint.set_pop(pop)

        if args.local_path:
            if not os.path.exists(args.local_path):
                log.warning(f"local path {args.local_path} does not exist")
                sys.exit(1)
            if os.path.isfile(args.local_path):
                local_files = [args.local_path]
            else:
                local_files = []

                for f in os.listdir(args.local_path):
                    local_file = os.path.join(args.local_path, f)
                    if os.path.isfile(local_file):
                        local_files.append(local_file)
            jobs = []
            async def on_ready(job: Job, path: str):
                nonlocal visualize_prediction
                nonlocal visualize_path
                while result := await job.predict():
                    visualize_prediction = result
                    visualize_path = path
                    if args.output:
                        print(path, json.dumps(result, indent=2))
            for local_file in local_files:
                job = await endpoint.upload(local_file, params=params, motion_detect=motion_detect, roi=args.roi, fps=args.fps)
                jobs.append(on_ready(job, local_file))
            await asyncio.gather(*jobs)
            if args.visualize and visualize_prediction is not None:
                image = Image.open(visualize_path)
                buffer = BytesIO()
                image.save(buffer, format="PNG")
                example_image_src = f"data:image/png;base64, {base64.b64encode(buffer.getvalue()).decode()}"
        elif args.url:
            job = await endpoint.load_from(args.url, params=params, motion_detect=motion_detect, roi=args.roi, fps=args.fps)
            while result := await job.predict():
                visualize_prediction = result
                if args.output:
                    print(json.dumps(result, indent=2))
            if args.visualize:
                example_image_src = args.url
        elif args.asset_uuid:
            job = await endpoint.load_asset(args.asset_uuid, params=params, motion_detect=motion_detect, roi=args.roi, fps=args.fps)
            while result := await job.predict():
                visualize_prediction = result
                if args.output:
                    print(json.dumps(result, indent=2))
            if args.visualize:
                async with EyePopSdk.dataEndpoint(is_async=True) as dataEndpoint:
                    buffer = await dataEndpoint.download_asset(
                        args.asset_uuid,
                        transcode_mode=TranscodeMode.image_original_size
                    ).read()
                    example_image_src = f"data:image/jpeg;base64, {base64.b64encode(buffer).decode()}"
    return visualize_prediction, example_image_src

visualize_result, example_image_src = asyncio.run(main(main_args))
if main_args.visualize:
    with open(os.path.join(script_dir, 'viewer.html')) as file:
        compiler = Compiler()
        html_template = compiler.compile(file.read())

    preview = html_template({
        'image_src': example_image_src,
        'result_json': json.dumps(visualize_result)
    })
    window = webui.window()
    window.set_root_folder('.')
    window.show(preview)
    webui.wait()



